from enum import Enum
from typing import TypeVar, Callable, Union

T = TypeVar('T')
K = TypeVar('K')


class ActionType(Enum):
    MAP = 1
    FILTER = 2
    REMOVE = 3
    FOREACH = 4
    REDUCE = 5
    FIND = 6
    COLLECT = 7


ArrayTransformer = Callable[[T], K]
FilterTransformer = Callable[[T], bool]
ForEachTransformer = Callable[[T], None]
ReduceTransformer = Callable[[K, T], K]
Condition = Union[Callable[[T], bool], T]


class StreamInterrupt:
    def __init__(self):
        self.interrupted = False


class Action:
    def __init__(self, type: ActionType, transformer: ArrayTransformer):
        self.type = type
        self.transformer = transformer


class __Arachnea:
    def __init__(self, lst: list):
        self.lst = lst
        self.action_stack = []

    def map(self, transformer: ArrayTransformer) -> 'Arachnea[K]':
        self.action_stack.append(Action(ActionType.MAP, transformer))
        return self

    def filter(self, transformer: FilterTransformer) -> 'Arachnea[T]':
        self.action_stack.append(Action(ActionType.FILTER, transformer))
        return self

    def forEach(self, transformer: ForEachTransformer) -> 'Arachnea[T]':
        self.action_stack.append(Action(ActionType.FOREACH, transformer))
        return self

    def remove(self, condition: Condition) -> 'Arachnea[T]':
        def condition_to_array_transformer(ele: T) -> bool:
            if callable(condition):
                return condition(ele)
            else:
                return ele == condition

        self.action_stack.append(Action(ActionType.REMOVE, condition_to_array_transformer))
        return self

    def action_loop(self, operation: Callable[[T], None], interrupt: StreamInterrupt = StreamInterrupt()):
        index = 0

        while index < len(self.lst):
            if interrupt.interrupted:
                break

            ele = self.lst[index]
            exclude = False

            for action in self.action_stack:
                if action.type == ActionType.MAP:
                    ele = action.transformer(ele)
                elif action.type == ActionType.FOREACH:
                    action.transformer(ele)
                elif action.type == ActionType.FILTER:
                    exclude = not action.transformer(ele)
                elif action.type == ActionType.REMOVE:
                    exclude = action.transformer(ele)

                    if exclude:
                        self.action_stack.remove(action)
                        break

                if exclude:
                    break

            if not exclude:
                operation(ele)

            index += 1

    def reduce(self, transformer: ReduceTransformer, initial_value: K) -> K:
        accumulator = initial_value

        def operation(ele: T):
            nonlocal accumulator
            accumulator = transformer(accumulator, ele)

        self.action_loop(operation)
        return accumulator

    def find(self, condition: Condition) -> T:
        def condition_to_array_transformer(ele: T) -> bool:
            if callable(condition):
                return condition(ele)
            else:
                return ele == condition

        result = None
        interrupt = StreamInterrupt()

        def operation(ele: T):
            nonlocal result
            if condition_to_array_transformer(ele):
                result = ele
                interrupt.interrupted = True

        self.action_loop(operation, interrupt)
        return result

    def collect(self) -> list:
        result = []

        def operation(ele: T):
            result.append(ele)

        self.action_loop(operation)
        return result


def arachnea(lst: list) -> __Arachnea:
    return __Arachnea(lst)



import unittest
from arachnea import arachnea

class TestArachnea(unittest.TestCase):
    arr = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

    def test_map_and_filter(self):
        # Mapping and filtering: Square the elements and filter odd numbers
        result = (
            arachnea(self.arr)
            .map(lambda ele: ele * ele)
            .filter(lambda ele: ele % 2 == 0)
            .collect()
        )
        expected = [4, 16, 36, 64, 100]
        self.assertEqual(result, expected)

    def test_reduce_to_sum(self):
        # Reduce to sum: Sum all elements
        result = (
            arachnea(self.arr)
            .reduce(lambda acc, ele: acc + ele, 0)
        )
        expected = sum(self.arr)
        self.assertEqual(result, expected)

    def test_find_element(self):
        # Find element: Find the first element greater than 5
        result = (
            arachnea(self.arr)
            .find(lambda ele: ele > 5)
        )
        expected = 6
        self.assertEqual(result, expected)

    def test_remove_element(self):
        # Remove element: Remove the first occurrence of 3
        result = (
            arachnea(self.arr)
            .remove(3)
            .collect()
        )
        expected = [1, 2, 4, 5, 6, 7, 8, 9, 10]
        self.assertEqual(result, expected)

    def test_for_each_operation(self):
        # ForEach operation: Print each element
        output = []
        (
            arachnea(self.arr)
            .forEach(lambda ele: output.append(ele))
            .collect()
        )
        expected = self.arr
        self.assertEqual(output, expected)

    def test_chaining_operations(self):
        # Chaining operations: Filter numbers greater than 5, double them, and sum
        result = (
            arachnea(self.arr)
            .filter(lambda ele: ele > 5)
            .map(lambda ele: ele * 2)
            .reduce(lambda acc, ele: acc + ele, 0)
        )
        expected = sum([ele * 2 for ele in self.arr if ele > 5])
        self.assertEqual(result, expected)

    def test_combining_operations(self):
        # Combining operations: Remove even numbers, square elements, and collect as strings
        result = (
            arachnea(self.arr)
            .filter(lambda ele: ele % 2 != 0)
            .map(lambda ele: ele * ele)
            .map(lambda ele: str(ele))
            .collect()
        )
        expected = [str(ele * ele) for ele in self.arr if ele % 2 != 0]
        self.assertEqual(result, expected)

    def test_error_cases(self):
        # Error cases: Test removing elements and handling an empty array
        result = (
            arachnea([])
            .remove(1)
            .collect()
        )
        expected = []
        self.assertEqual(result, expected)

if __name__ == '__main__':
    unittest.main()



# setup.py

from setuptools import setup, find_packages

setup(
    name='arachnea',
    version='0.0.2',
    packages=find_packages(),
    description='A Python library for efficient array operations using a fluent API.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Your Name',
    author_email='your.email@example.com',
    keywords='array operations, data processing, lightweight, Python library',
    url='https://github.com/yourusername/arachnea',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
)
